@@ -751,7 +751,8 @@ collect(Src, File, In, Options) :-
         Comments = []
     ;   CommentHandling == store
     ->  CommentOptions = [ process_comment(true) ],
-        Comments = []
+        Comments = [],
+	set_prolog_flag(xref_store_comments, true)
     ;   CommentOptions = [ comments(Comments) ]
     ),
     repeat,
@@ -769,7 +770,8 @@ collect(Src, File, In, Options) :-
                   E, print_message(error, E)),
             erase(Ref)),
         EOF == true,
-    !.
+    !,
+    set_prolog_flag(xref_store_comments, false).
 
 report_syntax_error(E, _, _) :-
     fatal_error(E),
@@ -1451,7 +1453,7 @@ hook(user:prolog_clause_name(_,_)).
 hook(user:prolog_list_goal(_)).
 hook(user:prolog_predicate_name(_,_)).
 hook(user:prolog_trace_interception(_,_,_,_)).
-hook(user:prolog_exception_hook(_,_,_,_)).
+hook(prolog:prolog_exception_hook(_,_,_,_,_)).
 hook(sandbox:safe_primitive(_)).
 hook(sandbox:safe_meta_predicate(_)).
 hook(sandbox:safe_meta(_,_)).
